"""
Aligns precipitation data and outputs a GSSHA file with a specified event spacing. All input gage files must be tab
delimited and are assumed to be YYYY\tMM\tDD\tHH\tMM\tValue format. The user must format the precipitation and perform
any quality control, preprocessing, or unit conversion. This script also assumes that if a gage does not record
precipitation at a given time, the value is a true zero rather than a false negative. This allows for all gages to
remain as active across all events to prevent degeneration into uniform or temporally invariant precipitation.

To use this script, the user should modify the entries in the user inputs section. Below are the steps for using this
script:

1) Specify the output name of the gage file to which the event series will be written
2) Specify the time increment of the input files. The value should be given in minutes.
3) Specify the spacing between events, in minutes, for the output file. The script looks for this dry duration between
   events in order to divide the period of record into events.
4) Specify the names of the input files. The input files should be in the same directory as the script.
5) Specify the type of gage. Valid gage types are:
    'incremental' - Standard precipitation gage which records the depth of precipitation deposited over a time
                    increment.
    'rate' - Precipitation that records the rate of precipitation being deposited over the time increment. This value
             should be given in depth/min.
    'tipping-bucket' - Not currently implmented
6) Specify the time step increment for rate and incremental gages in minutes. Tipping-bucket gages should have zero
   entered as there is not a uniform timestep.
7) Specify the units. Valid identifies are "mm" and "incheas"
8) Specify the easting and northing locations of the gages. These should be entered in the same order as the gages are
   listed in the input file list.
9) Run the script. If the input files and configuration are setup correctly, the gage file should be generated. It is
   recommended that the user run the gage file with GSSHA through the period of record to confirm that there are no 
   unanticipated issues.
   
Author: Drew Allan Loney, PhD PE
Email: drew.a.loney@usace.army.mil
Version: 1.0

It is requested that any errors found within the script be conveyed to the script author. These will be corrected
in a timely manner, with an updated script posted for the community's use.


This script is released under the MIT license:   
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE 
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import numpy as np
import datetime, warnings
from multiprocessing import Pool, freeze_support
from functools import partial

##### User inputs #####
s_output_filename = '2017_09_11_upper_russian_24.gag'
i_output_time_increment = 15 											# Minutes
i_output_event_spacing = 60 * 24											# Minutes
i_number_of_processors = 3                                              # Number of processors to utilize

sl_gage_files = ['p_kuki.txt', 'p_hky.txt', 'p_clv.txt', 'q_cdw.txt', 'p_boonville.txt',
                 '040212q_angwin_pac_union.txt', '040368q_atlas_road.txt', '041507q_capay_5.txt',
                 '041603q_cazadero_5.txt', '041806q_clearlake_4.txt', '041839q_cloverdale.txt',
                 '042084q_covelo_eel_ranger.txt', '043161q_fort_bragg_5.txt', '043194q_fort_seward.txt',
                 '044689q_lake_mendocino_dam.txt', '044712q_lake_solano.txt', '044851q_laytonville.txt',
                 '044883q_leggett.txt', '045258p_mahnke.txt', '045258q_mahnke.txt', '045711p_miranda.txt',
                 '045711q_miranda.txt', '045713p_miranda_spengler_ran.txt', '045713q_miranda_spengler_ran.txt',
                 '046105p_navarro.txt', '046105q_navarro.txt', '046290q_novato_8.txt', '046356q_oakville.txt',
                 '046726p_paskenta_ranger_station.txt', '046726q_paskenta_ranger_station.txt',
                 '046826q_petaluma_airport.txt', '047009p_point_arena.txt', '047009q_point_arena.txt',
                 '047107p_potter_valley_3_nnw.txt', '047108p_potter_valley_3_se.txt',
                 '047109p_potter_valley_powerhouse.txt', '047109q_potter_valley_powerhouse.txt',
                 '047351p_redwood_valley.txt', '047646q_saint_helena.txt', '047649p_saint_helena.txt',
                 '047971p_santa_rosa_airport.txt', '048072q_sebastopol.txt', '048587p_stony_gorge_reservoir.txt',
                 '048587q_stony_gorge_reservoir.txt', '048885p_the_geysers.txt', '048885q_the_geysers.txt',
                 '049129p_ukiah_masonite.txt', '049273p_venado.txt', '049273q_venado.txt',
                 '049440q_warm_springs_dam.txt', '049677p_williams.txt', '049677q_williams.txt',
                 '049678p_williams_airport.txt', '049684q_willits.txt', '049685p_howard_forest.txt',
                 '049685q_howard_forest.txt', '049700p_willows_usbr.txt', '049851p_yorkville.txt',
                 '049851q_yorkville.txt', '049953p_zenia_kettenpom.txt', '049953q_zenia_kettenpom.txt',
                 '736587p_finley_5.txt', '736597p_finley.txt', '736682p_hopland_8.txt', '736592p_kelseyville_4.txt',
                 '736595p_lakeport.txt']
sl_gage_types = ['incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental', 'incremental', 'incremental', 'incremental', 'incremental', 'incremental',
                 'incremental']

il_gage_increments = [60, 60, 60, 15, 60, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 60, 15, 60, 15, 60, 15,
                      60, 15, 15, 15, 60, 15, 15, 60, 15, 60, 60, 60, 15, 60, 15, 60, 60, 15, 60, 15, 60, 15, 60, 60,
                      15, 15, 60, 15, 60, 15, 60, 15, 60, 60, 15, 60, 15, 60, 60, 60, 60, 60]
s_units = 'mm'

dl_gage_easting_location = [482644.373, 513908.932, 495350.9521, 484001.6801, 469809.9321, 549367.02, 565459.07,
                            575327.99, 485842.97, 531765.53, 500000.00, 492869.71, 434956.40, 446110.21, 484177.40,
                            587198.88, 458563.02, 438720.99, 518801.39, 518801.39, 433313.99, 433313.99, 434748.28,
                            434748.28, 451046.21, 451046.21, 524830.54, 546569.62, 539902.60, 539902.60, 532090.43,
                            439314.42, 439300.22, 489949.69, 494251.39, 488514.54, 488514.54, 482746.84, 540692.73,
                            555174.34, 516549.29, 516008.07, 540075.96, 540075.96, 515919.47, 514472.79, 482722.35,
                            498548.76, 498548.76, 500000.00, 573447.88, 573447.88, 573499.83, 469871.41, 472714.57,
                            472714.57, 568750.78, 479767.13, 479767.13, 460262.13, 460262.13, 504332.7829, 511543.7348,
                            500001.1646, 510113.2997, 507213.4688]
dl_gage_northing_location = [4330758.86, 4286493.787, 4303388.735, 4339330.998, 4315446.675, 4268843.47, 4254161.98,
                             4287541.09, 4268456.34, 4313141.85, 4294582.75, 4409262.07, 4373547.66, 4451998.23,
                             4342685.99, 4261767.16, 4394572.66, 4415053.88, 4300153.44, 4300153.44, 4448399.52,
                             4448399.52, 4450237.44, 4450237.44, 4335425.29, 4335425.29, 4220646.34, 4250332.00,
                             4414912.32, 4414912.32, 4231767.53, 4307762.31, 4305912.42, 4357474.12, 4350070.88,
                             4357476.11, 4357476.11, 4346388.98, 4261396.59, 4268881.39, 4261732.15, 4250213.27,
                             4381616.59, 4381616.59, 4294598.71, 4294595.94, 4335291.44, 4274239.57, 4274239.57,
                             4285335.91, 4333766.44, 4333766.44, 4328217.64, 4363074.99, 4355665.44, 4355665.44,
                             4376269.03, 4305705.44, 4305705.44, 4446360.61, 4446360.61, 4313078.562, 4318634.374,
                             4318625.919, 4309384.1, 4320478.008]

####################### No additional user input after this line #######################################################
##### File read parallel functions
def _parallel_read(s_gage):
    print('\tLoading ' + s_gage)
    return np.loadtxt(s_gage, delimiter='\t')

##### Padding parallel functions #####
def _parallel_padding_incremental(o_datetime, dm_gage):
    i_index = np.argwhere(np.logical_and(np.logical_and(np.logical_and(o_datetime.year == dm_gage[:, 0], o_datetime.month == dm_gage[:, 1]),
                                         np.logical_and(o_datetime.day == dm_gage[:, 2], o_datetime.hour == dm_gage[:, 3])),
                                         o_datetime.minute == dm_gage[:, 4])).flatten()

    # Append the value into the array
    if len(i_index) == 0:
        return 0

    else:
        return dm_gage[i_index[0], 5]



##### Main function #####
# Using a main function is necessary for parallel calculations
if __name__ == "__main__":
    freeze_support()

    ### Data validation ###
    assert len(sl_gage_files) == len(sl_gage_types) == len(il_gage_increments) == len(dl_gage_easting_location) == len(dl_gage_northing_location)

    ### Open a parallel pool ###
    print('Opening processing pool with ' + str(i_number_of_processors) + ' processors...')
    o_pool = Pool(processes=i_number_of_processors)

    ### Load the precip files ###
    # Print the working message
    print('Loading precipitation files for processing...')

    # Load the files
    al_gage_records = o_pool.map(_parallel_read, sl_gage_files)


    ### Find the start and stop date for the event from across the gages ###
    print('Determining the time series start and stop dates...')

    # Format must be year, month, day, hour, minute in the columns
    o_start_date = None
    o_stop_date = None

    for dm_gage in al_gage_records:

        # Get the start date
        o_start_date_gage = datetime.datetime(year=int(dm_gage[0,0]), month=int(dm_gage[0,1]), day=int(dm_gage[0,2]),
                                              hour=int(dm_gage[0,3]), minute=int(dm_gage[0,4]))

        # Get the stop date
        i_stop_index = len(dm_gage) - 1
        o_stop_date_gage = datetime.datetime(year=int(dm_gage[i_stop_index,0]), month=int(dm_gage[i_stop_index,1]),
                                             day=int(dm_gage[i_stop_index,2]), hour=int(dm_gage[i_stop_index,3]),
                                             minute=int(dm_gage[i_stop_index,4]))

        # Compare and replace the dates
        if o_start_date is None:
            o_start_date = o_start_date_gage
        elif o_start_date_gage < o_start_date:
            o_start_date = o_start_date_gage

        if o_stop_date is None:
            o_stop_date = o_stop_date_gage
        elif o_stop_date_gage > o_stop_date:
            o_stop_date = o_stop_date_gage

    ### Create filled array series for each gage to prepare for interpolation ###
    print('Padding and converting gage records...')

    # This section converts all of the gages to accumulation
    l_gage_records_filled = []

    for i_entry_gage in range(0, len(al_gage_records), 1):
        # Print the current gage
        print('\tGage ' + str(i_entry_gage + 1) + ' of ' + str(len(al_gage_records)))

        # Extract the working gage
        dm_gage = al_gage_records[i_entry_gage]

        if sl_gage_types[i_entry_gage] == 'incremental':
            # Standard accumulative gage. This stores the depth of water deposited over a time increment.
            print('\t\tIncremental gage. Padding the time series with zero entries.')

            # Get the start date
            o_start_date_gage = datetime.datetime(year=int(dm_gage[0, 0]), month=int(dm_gage[0, 1]), day=int(dm_gage[0, 2]),
                                                  hour=int(dm_gage[0, 3]), minute=int(dm_gage[0, 4]))

            # Get the stop date
            i_stop_index = len(dm_gage) - 1
            o_stop_date_gage = datetime.datetime(year=int(dm_gage[len(dm_gage) - 1, 0]),
                                                 month=int(dm_gage[len(dm_gage) - 1, 1]),
                                                 day=int(dm_gage[len(dm_gage) - 1, 2]),
                                                 hour=int(dm_gage[len(dm_gage) - 1, 3]),
                                                 minute=int(dm_gage[len(dm_gage) - 1, 4]))

            # Create a timedelta between series values
            o_timedelta = datetime.timedelta(minutes=il_gage_increments[i_entry_gage])

            # Calculate the number of timesteps needed to construct the series
            i_number_of_increments = ((o_stop_date_gage - o_start_date_gage).total_seconds() / 60.0) / \
                                     float(il_gage_increments[i_entry_gage])

            # Construct a timestep list
            l_timesteps = [o_start_date_gage]

            o_running_date_gage = o_start_date_gage
            for i_entry_timestep in range(0, int(i_number_of_increments), 1):
                # Construct each of the timesteps
                o_running_date_gage += o_timedelta

                # Append into the list
                l_timesteps.append(o_running_date_gage)

            # Create the value list
            dl_values = o_pool.map(partial(_parallel_padding_incremental, dm_gage=dm_gage), l_timesteps)

            # Repackage the event into timestamp/value format
            dl_gage_updated = []

            for i_entry_timestep in range(0, int(i_number_of_increments), 1):
                dl_gage_updated.append([l_timesteps[i_entry_timestep], dl_values[i_entry_timestep]])

            # Append into the event
            l_gage_records_filled.append(dl_gage_updated)

        elif sl_gage_types[i_entry_gage] == 'rate':
            # Gage that records the precipitation rate over the time increment. This assumes that the deposition is in
            # units/min

            print('\t\tRate gage. Converting to incremental and padding the time series with zero entries.')

            # Get the start date
            o_start_date_gage = datetime.datetime(year=int(dm_gage[0, 0]), month=int(dm_gage[0, 1]), day=int(dm_gage[0, 2]),
                                                  hour=int(dm_gage[0, 3]), minute=int(dm_gage[0, 4]))

            # Get the stop date
            i_stop_index = len(dm_gage) - 1
            o_stop_date_gage = datetime.datetime(year=int(dm_gage[len(dm_gage) - 1, 0]),
                                                 month=int(dm_gage[len(dm_gage) - 1, 1]),
                                                 day=int(dm_gage[len(dm_gage) - 1, 2]),
                                                 hour=int(dm_gage[len(dm_gage) - 1, 3]),
                                                 minute=int(dm_gage[len(dm_gage) - 1, 4]))

            # Create a timedelta between series values
            o_timedelta = datetime.timedelta(minutes=il_gage_increments[i_entry_gage])

            # Calculate the number of timesteps needed to construct the series
            i_number_of_increments = ((o_stop_date_gage - o_start_date_gage).total_seconds() / 60.0) / \
                                     float(il_gage_increments[i_entry_gage])

            # Construct a timestep list
            l_timesteps = [o_start_date_gage]

            o_running_date_gage = o_start_date_gage
            for i_entry_timestep in range(0, int(i_number_of_increments), 1):
                # Construct each of the timesteps
                o_running_date_gage += o_timedelta

                # Append into the list
                l_timesteps.append(o_running_date_gage)

            # Create the value list
            dl_values = []

            for i_entry_timestep in range(0, int(i_number_of_increments), 1):
                i_index = np.argwhere(
                    np.logical_and(np.logical_and(np.logical_and(l_timesteps[i_entry_timestep].year == dm_gage[:, 0],
                                                                 l_timesteps[i_entry_timestep].month == dm_gage[:, 1]),
                                                  np.logical_and(l_timesteps[i_entry_timestep].day == dm_gage[:, 2],
                                                                 l_timesteps[i_entry_timestep].hour == dm_gage[:, 3])),
                                   l_timesteps[i_entry_timestep].minute == dm_gage[:, 4])).flatten()

                # Append the value into the array
                if len(i_index) == 0:
                    dl_values.append(0)

                else:
                    dl_values.append(dm_gage[i_index[0], 5] * float(il_gage_increments[i_entry_gage]))

            # Repackage the event into timestamp/value format
            dl_gage_updated = []

            for i_entry_timestep in range(0, int(i_number_of_increments), 1):
                dl_gage_updated.append([l_timesteps[i_entry_timestep], dl_values[i_entry_timestep]])

            # Append into the event
            l_gage_records_filled.append(dl_gage_updated)

        elif sl_gage_types[i_entry_gage] == 'tipping-bucket':
            # Tipping bucket gage. This records the number of tips per unit time.
            print('\t\tTipping bucket gage. Coverting to incremental, excluding tips based on event spacing, and padding '
                  'the time series with zero entries.')

            ## Difference the gage values to obtain the incremental precipitation ##
            da_values = np.zeros(len(dm_gage))
            for i_entry_timestep in range(1, len(dm_gage), 1):
                da_values[i_entry_timestep] = dm_gage[i_entry_timestep, 5] - dm_gage[i_entry_timestep - 1, 5]

            dm_gage[:, 5] = da_values

            ## Remove any timesteps with negative values ###
            ia_negative_indices = np.argwhere(dm_gage[:, 5] < 0).flatten()
            dm_gage = np.delete(dm_gage, ia_negative_indices, 0)

            ## Determine the minimum precipitation ##
            d_minimum_timestep = np.min(dm_gage[:, 5])

            if s_units == 'mm':
                d_minimum_timestep = 0.1
            elif s_units == 'inches':
                d_minimum_timestep = 0.01
            else:
                raise NotImplementedError('Precipitation unit is not understood.')

            ## Combine any entries occuring within the same minute##
            # Create the holding arrays
            ol_timesteps_combined = []
            dl_values_combined = []

            # Create the initial datetime and running total
            o_running_date = datetime.datetime(year=int(dm_gage[0, 0]), month=int(dm_gage[0, 1]), day=int(dm_gage[0, 2]),
                                              hour=int(dm_gage[0, 3]), minute=int(dm_gage[0, 4]))
            d_running_total = dm_gage[0, 5]

            # Loop over the list
            for i_entry_timestep in range(1, len(dm_gage), 1):
                # Construct the new datatime objection
                o_data_test = datetime.datetime(year=int(dm_gage[i_entry_timestep, 0]),
                                                month=int(dm_gage[i_entry_timestep, 1]),
                                                day=int(dm_gage[i_entry_timestep, 2]),
                                                hour=int(dm_gage[i_entry_timestep, 3]),
                                                minute=int(dm_gage[i_entry_timestep, 4]))

                if o_data_test == o_running_date:
                    # Timesteps are the same. Multiple tips in a minute. Add to the running total.
                    d_running_total += dm_gage[i_entry_timestep, 5]

                else:
                    # Timesteps are different. Append the previous time and and set the current value as the running
                    ol_timesteps_combined.append(o_running_date)
                    dl_values_combined.append(d_running_total)

                    o_running_date = o_data_test
                    d_running_total = dm_gage[i_entry_timestep, 5]

            # Append the last set of values into the list
            ol_timesteps_combined.append(o_running_date)
            dl_values_combined.append(d_running_total)

            ## Remove isolated tips that aren't culustered within the output event spacing ##
            ia_indices2delete = []

            for i_entry in range(0, len(ol_timesteps_combined), 1):
                # Look backward for the event window
                b_backward = True
                if i_entry > 0:
                    i_time_difference = (ol_timesteps_combined[i_entry] - ol_timesteps_combined[i_entry - 1]).total_seconds() / 60.0

                    if i_time_difference > i_output_event_spacing and np.round(dl_values_combined[i_entry], 5) <= d_minimum_timestep:
                        b_backward = False

                # Look forward for the event window
                b_forward = True
                if i_entry < len(ol_timesteps_combined) - 1:
                    i_time_difference = (ol_timesteps_combined[i_entry + 1] - ol_timesteps_combined[i_entry]).total_seconds() / 60.0

                    if i_time_difference > i_output_event_spacing and np.round(dl_values_combined[i_entry], decimals=5) <= d_minimum_timestep:
                        b_forward = False

                # Add to the delete list if window is not satisfied
                if not b_backward and not b_forward:
                    ia_indices2delete.append(i_entry)

            # Promote the list to an array
            ia_indices2delete = np.array(ia_indices2delete)

            # Remove the entries from each of the lists
            ol_timesteps_filtered = [ol_timesteps_combined[x] for x in range(0, len(ol_timesteps_combined), 1)
                                     if x not in ia_indices2delete]
            dl_values_filtered = [dl_values_combined[x] for x in range(0, len(dl_values_combined), 1)
                                  if x not in ia_indices2delete]

            ## Create the output series based on the output increment ##
            o_output_timedelta = datetime.timedelta(minutes=i_output_time_increment)
            i_number_of_increments = int(np.ceil((o_stop_date - o_start_date).total_seconds() / 60.0 / i_output_time_increment))

            ol_gage_output_timeseries = [o_start_date]

            o_running_date = o_start_date
            for i_entry in range(0, i_number_of_increments, 1):
                o_running_date += o_output_timedelta
                ol_gage_output_timeseries.append(o_running_date)

            ## Aggregate the incremental values onto the new series ##
            # Promote the value list to an array to allow for indexing
            da_values_filtered = np.array(dl_values_filtered)
            del dl_values_filtered

            # Create a list to hold the data
            ol_gage_output_values = []

            # Loop over every measurement in the output series
            for i_entry_output_measurement in range(0, len(ol_gage_output_timeseries), 1):
                # Calculate the time difference between the current entry and all times in the date series
                ia_time_difference = np.array([(ol_gage_output_timeseries[i_entry_output_measurement] - ol_timesteps_filtered[x]).total_seconds() / 60.0
                                               for x in range(0, len(ol_timesteps_filtered), 1)])

                # Find entries that are positive and within the output window
                ia_valid_indices = np.argwhere(np.logical_and(ia_time_difference >= 0,
                                                              ia_time_difference <= i_output_time_increment))

                # Sum the contributing values
                if len(ia_valid_indices) > 0:
                    # There are entries that contribute to the timestep
                    d_value = np.sum(da_values_filtered[ia_valid_indices])

                else:
                    # There are no events that contribute to the the current timestep. Value is zero
                    d_value = 0

                # Append the value into the new list
                ol_gage_output_values.append(d_value)

            ## Repackage the event into timestamp/value format ##
            dl_gage_updated = []

            for i_entry_timestep in range(0, int(i_number_of_increments), 1):
                dl_gage_updated.append([ol_gage_output_timeseries[i_entry_timestep],
                                        ol_gage_output_values[i_entry_timestep]])

            ## Remove the start and end of the series to remove artifical zeros ##
            i_start_index = -1
            for i_entry in range(0, len(dl_gage_updated), 1):
                if dl_gage_updated[i_entry][1] == 0:
                    i_start_index = i_entry
                else:
                    break

            i_stop_index = -1
            for i_entry in range(len(dl_gage_updated)-1, -1, -1):
                if dl_gage_updated[i_entry][1] == 0:
                    i_stop_index = i_entry
                else:
                    break

            dl_gage_updated = [dl_gage_updated[x] for x in range(i_start_index, i_stop_index+1, 1)]

            # Append into the event
            l_gage_records_filled.append(dl_gage_updated)

            # Update the interval list value to account for tipping-bucket series processing
            il_gage_increments[i_entry_gage] = i_output_time_increment

        else:
            # The gage type is not recognized. Throw an error.
            raise AttributeError('Gage type note recognized for gage number ' + str(i_entry_gage))

    ### Interpolate series onto the correct output increment ###
    print('Interpolating time series onto the output time timeseries...')
    ## Construct the output time series ##
    o_output_timedelta = datetime.timedelta(minutes=i_output_time_increment)

    # Calculate the number of timesteps needed to construct the series
    i_number_of_increments = int(np.ceil(((o_stop_date - o_start_date).total_seconds() / 60.0) / i_output_time_increment))

    # Create the timeseries list
    l_timeseries_output = [o_start_date]

    # Construct the timeseries
    o_running_date = o_start_date
    for i_entry_timestep in range(0, i_number_of_increments, 1):
        o_running_date += o_output_timedelta
        l_timeseries_output.append(o_running_date)

    ## Move the gages onto the new timeseries ##
    dm_gages_interpolated = np.zeros((len(l_timeseries_output), len(sl_gage_files)))

    for i_entry_gage in range(0, len(l_gage_records_filled), 1):
        # Gages have already been transferred into accumulated equivalent. Dissagregate or aggregate onto the output
        # time series

        # Find the analysis mode for the event
        if il_gage_increments[i_entry_gage] > i_output_time_increment:
            # Check for disaggregation
            s_mode = 'disaggregate'
            print('\tGage ' + str(i_entry_gage + 1) + ' of ' + str(len(l_gage_records_filled)) + '. Disaggregate.')

        elif il_gage_increments[i_entry_gage] < i_output_time_increment:
            # Check for aggregation
            s_mode = 'aggregate'
            print('\tGage ' + str(i_entry_gage + 1) + ' of ' + str(len(l_gage_records_filled)) + '. Aggregate.')

        elif il_gage_increments[i_entry_gage] == i_output_time_increment:
            # Check for the same increment
            s_mode = 'maintain'
            print('\tGage ' + str(i_entry_gage + 1) + ' of ' + str(len(l_gage_records_filled)) + '. Skipping interpolation.')

        else:
            # Not understood. Error out.
            raise NotImplementedError('Aggregation/disaggregation mode is not understood')

        # Act on the analysis mode to move onto the output timeseries
        if s_mode == 'disaggregate':
            # Disaggregate the event timesteps onto the shorter duration output timeseries
            # Calculate the number of gage intervals that each entry gets divided into
            i_interval_spacing = il_gage_increments[i_entry_gage] / i_output_time_increment

            # Check that the interval is an integer. Noninteger intervals are not handled.
            if i_interval_spacing % 1 > 0:
                # Noninteger interval spacing. This has not been handled. Raise an error.
                raise NotImplementedError('Precipitation aggregation for noninteger interval spacing has not been '
                                          'implemented.')

            # Find the starting time index
            i_index = [x for x in range(0, len(l_timeseries_output), 1) if
                       l_timeseries_output[x] == l_gage_records_filled[i_entry_gage][0][0]]

            if len(i_index) > 0:
                # Exact match of the starting index has been found
                i_index = i_index[0]

            else:
                # Take the first output index before the gage index
                i_index = [x for x in range(0, len(l_timeseries_output), 1) if
                           l_timeseries_output[x] < l_gage_records_filled[i_entry_gage][0][0]]

                # Promote the an array and take the maximum value
                i_index = np.max(np.array(i_index))

            # From the index, disaggregate and fill the array
            i_counter = 0
            for i_entry_measurement in range(0, len(l_gage_records_filled[i_entry_gage]), 1):
                for i_entry_submeasurement in range(0, int(i_interval_spacing), 1):
                    if i_index + i_counter < len(l_timeseries_output) - 1:
                        dm_gages_interpolated[i_index + i_counter, i_entry_gage] = \
                            l_gage_records_filled[i_entry_gage][i_entry_measurement][1] / float(i_interval_spacing)

                    # Increment the counter
                    i_counter += 1

            del i_counter

        elif s_mode == 'aggregate':
            # Aggregate the event timesteps onto the longer duration output timeseries
            # Calculate the number of gage intervals that get combined into a single output interval
            i_interval_spacing = i_output_time_increment / il_gage_increments[i_entry_gage]

            # Check that the interval is an integer. Noninteger intervals are not handled.
            if i_interval_spacing % 1 > 0:
                # Noninteger interval spacing. This has not been handled. Raise an error.
                raise NotImplementedError('Precipitation aggregation for noninteger interval spacing has not been '
                                          'implemented.')

            # Find the starting time index
            i_index = [x for x in range(0, len(l_timeseries_output), 1) if
                       l_timeseries_output[x] == l_gage_records_filled[i_entry_gage][0][0]]

            if len(i_index) > 0:
                # Exact match of the starting index has been found
                i_index = i_index[0]

            else:
                # Take the first output index before the gage index
                i_index = [x for x in range(0, len(l_timeseries_output), 1) if
                           l_timeseries_output[x] < l_gage_records_filled[i_entry_gage][0][0]]

                # Promote the an array and take the maximum value
                i_index = np.max(np.array(i_index))

            # From the index, aggregate and fill the array
            i_counter = 0
            for i_entry_measurement in range(0, len(l_gage_records_filled[i_entry_gage]), int(i_interval_spacing)):
                for i_entry_submeasurement in range(i_entry_measurement, i_entry_measurement + int(i_interval_spacing), 1):
                    if i_entry_measurement < len(l_timeseries_output) - 1:
                        dm_gages_interpolated[i_index + i_counter, i_entry_gage] += \
                            l_gage_records_filled[i_entry_gage][i_entry_submeasurement][1]

                # Increment the counter
                i_counter += 1

            del i_counter

        elif s_mode == 'maintain':
            # Maintain the timestep on the series.
            # No interpolation is necessary. Find the starting point and write the data directly into the list.
            i_index = [x for x in range(0, len(l_timeseries_output), 1) if l_timeseries_output[x] == l_gage_records_filled[i_entry_gage][0][0]]
            i_index = i_index[0]

            # Beginning from the index, lay in the data sequentially
            for i_entry_measurement in range(0, len(l_gage_records_filled[i_entry_gage]), 1):
                dm_gages_interpolated[i_index + i_entry_measurement, i_entry_gage] = \
                    l_gage_records_filled[i_entry_gage][i_entry_measurement][1]

    ### Create the list filled with the time values ###
    # Create the empty array of the appropriate length
    dm_gages_interpolated = np.hstack((np.zeros((len(l_timeseries_output), 5)), dm_gages_interpolated))

    # Loop and fill the array
    for i_entry in range(0, len(l_timeseries_output), 1):
        # Put the time data into the table
        dm_gages_interpolated[i_entry, 0] = l_timeseries_output[i_entry].year
        dm_gages_interpolated[i_entry, 1] = l_timeseries_output[i_entry].month
        dm_gages_interpolated[i_entry, 2] = l_timeseries_output[i_entry].day
        dm_gages_interpolated[i_entry, 3] = l_timeseries_output[i_entry].hour
        dm_gages_interpolated[i_entry, 4] = l_timeseries_output[i_entry].minute


    ### Break up the events into the desired event length ###
    print('Separating the gage record into events...')
    # Calculate the number of increments between events
    i_spacing_increments = i_output_event_spacing / i_output_time_increment

    # Define an event holder matrix
    al_event_holder = []

    # Loop on the aligned data
    i_precip_start = None
    i_precip_end = None
    i_precip_dry = None


    for i_entry_row in range(0, len(dm_gages_interpolated), 1):
        # Loop on each row and count the number of entries

        if np.any(dm_gages_interpolated[i_entry_row, 5:]) > 0:
            # There is precipitation in the row
            if i_entry_row == len(dm_gages_interpolated) - 1:
                # Last entry in the file. Set the counters, segment the data, and clear the counters.
                i_precip_end = i_entry_row

                # Segment out the event
                al_event_holder.append(dm_gages_interpolated[i_precip_start:i_precip_end, :])

                # Reset the counters
                i_precip_start = None
                i_precip_end = None
                i_precip_dry = None

            elif i_precip_start is None:
                # No previous wet lines in the event. Indicate that this is the event start
                i_precip_start = i_entry_row

            else :
                # Check if the dry counters need to be reset
                if i_precip_end is not None:
                    # A wet entry has occured within the event spacing interval. Reset the counters to include it in the
                    # previous event.
                    i_precip_end = None
                    i_precip_dry = None

        else:
            # There is no precipitation in the row
            if i_entry_row == len(dm_gages_interpolated) - 1:
                # Last entry in the matrix. Determine if there is an event to append
                if i_precip_start:
                    # Event exists to append to the event list
                    if i_precip_end is None:
                        i_precip_end = i_entry_row

                    # Segment out the event
                    al_event_holder.append(dm_gages_interpolated[i_precip_start:i_precip_end, :])

                    # Reset the counters
                    i_precip_start = None
                    i_precip_end = None
                    i_precip_dry = None


            elif i_precip_start is not None and i_precip_end is None:
                # This if the first dry line after the event import is on. Enter the dry row number and begin counting
                # the lines to possibly break the evnet.
                i_precip_end = i_entry_row
                i_precip_dry = 1

            elif i_precip_start is not None and i_precip_end is not None:
                # This is a continuing dry line after the end of an event. Increment the counter and check if splitting the
                # event is required.
                i_precip_dry += 1

                if i_precip_dry == i_spacing_increments:
                    # There is the correct number of dry lines after the finishing of an event. Segment out the event and
                    # reset the counters
                    al_event_holder.append(dm_gages_interpolated[i_precip_start:i_precip_end, :])

                    # Reset the counters
                    i_precip_start = None
                    i_precip_end = None
                    i_precip_dry = None

    ### Flag entries that are outside of the gage active times ###
    al_event_holder_adjusted = []

    for i_entry_event in range(0, len(al_event_holder), 1):
        # Extract the active event
        dm_event_active = al_event_holder[i_entry_event]

        # Get the start and stop index for get gage
        for i_entry_gage in range(0, len(al_gage_records), 1):

            # Get the start date
            o_start_date_gage = datetime.datetime(year=int(al_gage_records[i_entry_gage][0,0]),
                                                  month=int(al_gage_records[i_entry_gage][0,1]),
                                                  day=int(al_gage_records[i_entry_gage][0,2]),
                                                  hour=int(al_gage_records[i_entry_gage][0,3]),
                                                  minute=int(al_gage_records[i_entry_gage][0,4]))

            # Get the stop date
            i_stop_index = len(al_gage_records[i_entry_gage]) - 1
            o_stop_date_gage = datetime.datetime(year=int(al_gage_records[i_entry_gage][i_stop_index,0]),
                                                 month=int(al_gage_records[i_entry_gage][i_stop_index,1]),
                                                 day=int(al_gage_records[i_entry_gage][i_stop_index,2]),
                                                 hour=int(al_gage_records[i_entry_gage][i_stop_index,3]),
                                                 minute=int(al_gage_records[i_entry_gage][i_stop_index,4]))

            # Loop on the event entries and adjust to NaN as needed
            for i_entry_measurement in range(0, len(dm_event_active), 1):
                o_measurement_date = datetime.datetime(year=int(dm_event_active[i_entry_measurement, 0]),
                                                       month=int(dm_event_active[i_entry_measurement, 1]),
                                                       day=int(dm_event_active[i_entry_measurement, 2]),
                                                       hour=int(dm_event_active[i_entry_measurement, 3]),
                                                       minute=int(dm_event_active[i_entry_measurement, 4]))

                if o_measurement_date < o_start_date_gage or o_measurement_date > o_stop_date_gage:
                    dm_event_active[i_entry_measurement, 5 + i_entry_gage] = np.NaN

        al_event_holder_adjusted.append(dm_event_active)


    ### Write the events to file ###
    print('Writing the final gage record...')

    # Open the log file
    o_output_file = open(s_output_filename, 'w')

    for i_entry_event in range(0, len(al_event_holder), 1):

        # Extract the event
        dm_event_active = al_event_holder_adjusted[i_entry_event]

        # Determine which gages are activate
        ia_active_gages = []
        for i_entry in range(0, len(sl_gage_files), 1):
            # Check which data columns are not completely filled with NaNs
            i_number_of_zeros = len(np.argwhere(np.isnan(dm_event_active[:, 5 + i_entry])).flatten())

            if i_number_of_zeros < len(dm_event_active):
                ia_active_gages.append(i_entry)

        # Promote the active gage list
        ia_active_gages = np.array(ia_active_gages)

        # Warn if only one gage in an event
        if len(ia_active_gages) == 1:
            warnings.warn('Single gage event. This is not compatible with long-term simulation mode.')

        # Print the event number
        o_output_file.write('EVENT \"Event ' + str(i_entry_event + 1) + '\"\n')

        # Print the number of records in the file
        o_output_file.write('NRPDS ' + str(len(dm_event_active) + 2) + '\n')

        # Print the number of active gages
        o_output_file.write('NRGAG ' + str(len(ia_active_gages)) + '\n')

        # Print the coordinates for each of the gages
        for i_entry_gage in range(0, len(ia_active_gages), 1):
            o_output_file.write('COORD ' + str("{:.1f}".format(dl_gage_easting_location[ia_active_gages[i_entry_gage]])) + ' ' +
                                str("{:.1f}".format(dl_gage_northing_location[ia_active_gages[i_entry_gage]])) + '\n')

        # Add a zero line at the start of the event
        # Construct a date from the first event line
        o_event_date = datetime.datetime(year=int(dm_event_active[0, 0]),
                                         month=int(dm_event_active[0, 1]),
                                         day=int(dm_event_active[0, 2]),
                                         hour=int(dm_event_active[0, 3]),
                                         minute=int(dm_event_active[0, 4]))

        # Decrement the date time by the output specified
        o_event_date -= datetime.timedelta(minutes=i_output_time_increment)

        # Create the array
        da_extra_entry = np.zeros(5 + len(sl_gage_files))
        da_extra_entry[0] = o_event_date.year
        da_extra_entry[1] = o_event_date.month
        da_extra_entry[2] = o_event_date.day
        da_extra_entry[3] = o_event_date.hour
        da_extra_entry[4] = o_event_date.minute

        # Extract the active event
        dm_event_active = np.vstack((da_extra_entry, dm_event_active))

        # Create an extra final line
        # Construct a date from the last event line
        o_event_date = datetime.datetime(year=int(dm_event_active[len(dm_event_active) - 1, 0]),
                                         month=int(dm_event_active[len(dm_event_active) - 1, 1]),
                                         day=int(dm_event_active[len(dm_event_active) - 1, 2]),
                                         hour=int(dm_event_active[len(dm_event_active) - 1, 3]),
                                         minute=int(dm_event_active[len(dm_event_active) - 1, 4]))

        # Decrement the date time by the output specified
        o_event_date += datetime.timedelta(minutes=i_output_time_increment)

        # Create the array
        da_extra_entry = np.zeros(5 + len(sl_gage_files))
        da_extra_entry[0] = o_event_date.year
        da_extra_entry[1] = o_event_date.month
        da_extra_entry[2] = o_event_date.day
        da_extra_entry[3] = o_event_date.hour
        da_extra_entry[4] = o_event_date.minute

        # Add it to the end of the event
        dm_event_active = np.vstack((dm_event_active, da_extra_entry))

        # Loop and output each line
        for i_entry_row in range(0, len(dm_event_active), 1):
            # Output the date, beginning with the year
            s_date_line = 'GAGES ' + str(int(dm_event_active[i_entry_row, 0])) + ' '

            # Correct month output
            if int(dm_event_active[i_entry_row, 1]) < 10:
                s_date_line = s_date_line + '0' + str(int(dm_event_active[i_entry_row, 1])) + ' '
            else:
                s_date_line = s_date_line + str(int(dm_event_active[i_entry_row, 1])) + ' '

            # Correct the day output
            if int(dm_event_active[i_entry_row, 2]) < 10:
                s_date_line = s_date_line + '0' + str(int(dm_event_active[i_entry_row, 2])) + ' '
            else:
                s_date_line = s_date_line + str(int(dm_event_active[i_entry_row, 2])) + ' '

            # Correct the hour output
            if int(dm_event_active[i_entry_row, 3]) < 10:
                s_date_line = s_date_line + '0' + str(int(dm_event_active[i_entry_row, 3])) + ' '
            else:
                s_date_line = s_date_line + str(int(dm_event_active[i_entry_row, 3])) + ' '

            # Correct the minute output
            if int(dm_event_active[i_entry_row, 4]) < 10:
                s_date_line = s_date_line + '0' + str(int(dm_event_active[i_entry_row, 4])) + ' '
            else:
                s_date_line = s_date_line + str(int(dm_event_active[i_entry_row, 4])) + ' '

            o_output_file.write(s_date_line)

            # Output the active gages
            for i_entry_gage in range(0, len(ia_active_gages), 1):
                # Output the value
                d_value = np.round(dm_event_active[i_entry_row, ia_active_gages[i_entry_gage] + 5], 4)

                if d_value > 0:
                    o_output_file.write(str("{:.4f}".format(d_value)))
                else:
                    o_output_file.write('0.0000')

                # Handle whether a tab or new line should be included
                if i_entry_gage == len(ia_active_gages) - 1:
                    # Last entry in the line. Write a new line break
                    o_output_file.write('\n')

                else:
                    # Not the last line. Write a tab.
                    o_output_file.write(' ')

        # Event has ended. Write a new line break.
        o_output_file.write('\n')

    # All events have been written. Close the file.
    o_output_file.flush()
    o_output_file.close()
